import _extends from "@babel/runtime/helpers/esm/extends";
import produce from "immer";
import { catchError, concatMap, delay, exhaustMap, filter, from, iif, map, mapTo, merge, of, retryWhen, Subject, switchMap, take, takeUntil, tap, throwError } from "rxjs"; // https://stackoverflow.com/questions/44911251/how-to-create-an-rxjs-retrywhen-with-delay-and-limit-on-tries

import { ModuleUtils } from "../../utils/lang";

const retryProcessing = config => {
  const retryPipeline = retryWhen(errors => errors.pipe( // Use concat map to keep the errors in order and make sure they
  // aren't executed in parallel
  concatMap((e, i) => // Executes a conditional Observable depending on the result
  // of the first argument
  iif(() => i > 10, // If the condition is true we throw the error (the last error)
  throwError(e), // Otherwise we pipe this back into our stream and delay the retry
  of(e).pipe(delay(500))))));
  return retryPipeline;
};

export class EntityFormClass {
  constructor(props, setState) {
    this.props = props;
    this.setState = setState;
    this.loadResource$ = new Subject();
    this.destroy$ = new Subject();
    this.reloadResource$ = new Subject();
    this.triggerSave$ = new Subject();
    this.triggerLoad$ = new Subject();
    this.newState = {
      loadStatus: {},
      params: {},
      resourceParams: {},
      resources: {},
      saveStatus: {},
      submitted: false
    };
    this.reloadResource = this.reloadResource.bind(this);
    this.configSave = this.configSave.bind(this);
    this.loadResourceConfig = this.loadResourceConfig.bind(this);
  }

  loadResourceConfig() {
    const processingStatus$ = new Subject();
    merge(processingStatus$.pipe(map(config => ({
      status: 'PROCESSING',
      config
    }))), this.loadResource$.pipe(tap(config => {
      processingStatus$.next(config);
    }), switchMap(config => {
      const data$ = config.config.data$ ? config.config.data$(config.params) : of(config.params);
      const loadData$ = ModuleUtils.isObservable(data$) ? data$ : ModuleUtils.isPromise(data$) ? from(data$) : of(data$);
      const retryConfig = config.config.retry ? config.config.retry : {
        interval: 100
      };
      return loadData$.pipe(retryProcessing(retryConfig), take(1), map(data => {
        return {
          status: 'PROCESSED',
          data,
          config
        };
      }), catchError(error => {
        return of({
          status: 'ERROR',
          error,
          config
        });
      }));
    }))).pipe(takeUntil(this.destroy$), tap(processingStatus => {
      processingStatus.config.config.onProcessingStatus(processingStatus); // processingStatus.data&&processingStatus.data.config&&processingStatus.data.config.config&&processingStatus.data.config.config.onProcessingStatus(processingStatus)

      this.setState(prevState => {
        const newState = produce(draft => {
          const config = processingStatus.config;
          draft.resources[config.id] = {
            status: processingStatus.status,
            data: processingStatus.data,
            error: processingStatus.error
          };
        })(prevState);
        this.newState = newState;
        return newState;
      });
    })).subscribe();
  }

  init() {
    this.loadResourceConfig();
    this.configSave();
    const loadConfig = this.props.load;

    if (!loadConfig) {
      this.setState(prevState => {
        const newState = produce(draft => {
          draft.loadStatus.status = 'PROCESSED';
          draft.loadStatus.data = {};
        })(prevState);
        this.newState = newState;
        return newState;
      });
      return;
    }

    const resources = loadConfig.resources || {};
    setTimeout(() => Object.keys(resources).map(key => _extends({}, resources[key], {
      key
    })).forEach(resource => {
      this.setState(prevState => {
        const newState = produce(draft => {
          draft.resources[resource.key] = {};
        })(prevState);
        this.newState = newState;
        return newState;
      });
      const processingStatus$ = new Subject();
      merge(processingStatus$.pipe(mapTo({
        status: 'PROCESSING'
      })), merge(this.reloadResource$.pipe(filter(reload => reload.id === resource.key), map(reload => reload.params)), resource.params$) // resource.params$
      .pipe(takeUntil(this.destroy$), tap(params => {
        this.setState(prevState => {
          const newState = produce(draft => {
            draft.resourceParams[resource.key] = params;
          })(prevState);
          this.newState = newState;
          return newState;
        });
        processingStatus$.next();
      }), switchMap(params => {
        const data$ = resource.data$ ? resource.data$(params) : of(params);
        const loadData$ = ModuleUtils.isObservable(data$) ? data$ : ModuleUtils.isPromise(data$) ? from(data$) : of(data$);
        const retryConfig = resource.retry ? resource.retry : {
          interval: 1000
        };
        return loadData$.pipe(retryProcessing(retryConfig), take(1), map(data => {
          return {
            status: 'PROCESSED',
            data
          };
        }), catchError(error => {
          return of({
            status: 'ERROR',
            error
          });
        }));
      }))).pipe(takeUntil(this.destroy$), tap(processingStatus => {
        resource.onProcessingStatus && resource.onProcessingStatus(processingStatus, {});
        this.setState(prevState => {
          const newState = produce(draft => {
            draft.resources[resource.key].status = processingStatus.status;
            draft.resources[resource.key].data = processingStatus.data;
            draft.resources[resource.key].error = processingStatus.error;
          })(prevState);
          this.newState = newState;
          return newState;
        });
      })).subscribe();
    }), 1);
    /******************************************************************************************************* */

    const adaptLoadData = loadConfig.entity.adaptData ? loadConfig.entity.adaptData : data => ({
      initialData: data,
      resources: {}
    });
    const processingStatus$ = new Subject();
    const params$ = this.props.params$ ? merge(this.props.params$, this.triggerLoad$) : this.triggerLoad$;
    merge(processingStatus$.pipe(mapTo({
      status: 'PROCESSING'
    })), params$.pipe(takeUntil(this.destroy$), tap(params => {
      this.setState(prevState => {
        const newState = produce(draft => {
          draft.params = params;
        })(prevState);
        this.newState = newState;
        return newState;
      });
      processingStatus$.next();
    }), switchMap(params => {
      const loadConfig = this.props.load ? this.props.load : {
        entity: {}
      };
      const retryConfig = loadConfig.entity.retry ? loadConfig.entity.retry : {
        interval: 200
      };
      const data$ = loadConfig.entity.data$ ? loadConfig.entity.data$(params) : of(params);
      const loadData$ = ModuleUtils.isObservable(data$) ? data$ : ModuleUtils.isPromise(data$) ? from(data$) : of(data$);
      return loadData$.pipe(retryProcessing(retryConfig), take(1), map((data, e) => {
        return adaptLoadData(data);
      }), map(data => {
        return {
          status: 'PROCESSED',
          data
        };
      }), catchError(error => {
        return of({
          status: 'ERROR',
          error
        });
      }));
    }))).pipe(takeUntil(this.destroy$), tap(processingStatus => {
      loadConfig.entity.onProcessingStatus && loadConfig.entity.onProcessingStatus(processingStatus, this.newState.params);

      if (processingStatus.status === 'PROCESSED') {
        this.setState(prevState => {
          const newState = produce(draft => {
            const resources = processingStatus.data.resources;
            Object.keys(resources).map(key => ({
              key,
              data: resources[key]
            })).forEach(item => {
              draft.resources[item.key] = {
                status: 'PROCESSED',
                data: item.data
              };
            });
          })(prevState);
          this.newState = newState;
          return newState;
        });
      }

      this.setState(prevState => {
        const newState = produce(draft => {
          var _processingStatus$dat;

          draft.loadStatus.status = processingStatus.status;
          draft.loadStatus.data = (_processingStatus$dat = processingStatus.data) == null ? void 0 : _processingStatus$dat.initialData;
          draft.loadStatus.error = processingStatus.error;
        })(prevState);
        this.newState = newState;
        return newState;
      });
    })).subscribe();
    /**************************************************************** */
  }

  configLoad() {
    const processingStatus$ = new Subject();
    this.triggerLoad$.pipe();
  }

  configSave() {
    const saveConfig = this.props.save;
    const processingStatus$ = new Subject();
    merge(processingStatus$.pipe(mapTo({
      status: 'PROCESSING'
    })), this.triggerSave$.pipe(tap(x => {
      processingStatus$.next();
    }), exhaustMap(data => {
      const data$ = saveConfig.saveDataFn$(data, this.newState.params);
      const saveData$ = ModuleUtils.isObservable(data$) ? data$ : ModuleUtils.isPromise(data$) ? from(data$) : of(data$);
      return saveData$.pipe(map(resp => ({
        status: 'PROCESSED',
        data: resp
      })), catchError(error => of({
        status: 'ERROR',
        error
      })));
    }))).pipe(tap(processingStatus => {
      saveConfig && saveConfig.onProcessingStatus && saveConfig.onProcessingStatus(processingStatus, this.newState.params);
      this.setState(prevState => {
        const newState = produce(draft => {
          draft.saveStatus.status = processingStatus.status;
          draft.saveStatus.data = processingStatus.data;
          draft.saveStatus.error = processingStatus.error;
        })(prevState);
        this.newState = newState;
        return newState;
      });
    })).subscribe();
  }

  close() {
    this.destroy$.next();
    this.destroy$.complete();
  }

  submit() {
    this.setState(prevState => {
      const newState = produce(draft => {
        draft.submitted = true;
      })(prevState);
      this.newState = newState;
      return newState;
    });
  }

  resetSubmit() {
    this.setState(prevState => {
      const newState = produce(draft => {
        draft.submitted = false;
      })(prevState);
      this.newState = newState;
      return newState;
    });
  }

  resetForm() {// this.setState(
    //     (prevState) => {
    //         const newState = produce(
    //             (draft: IEntityFormState) => {
    //                 draft.data = this.newState.loadStatus.data;
    //             }
    //         )(prevState)
    //         this.newState = newState;
    //         return newState;
    //     }
    // )
  }

  get formState() {
    return this.newState;
  }

  get resources() {
    return this.newState.resources;
  }

  reloadResource(resourceId) {
    this.reloadResource$.next({
      id: resourceId,
      params: this.newState.resourceParams[resourceId]
    });
  }

  get config() {
    var _this$formState;

    return {
      formState: this.formState,
      resources: (_this$formState = this.formState) == null ? void 0 : _this$formState.resources,
      reloadResource: this.reloadResource,
      submit: this.submit.bind(this),
      resetSubmit: this.resetSubmit,
      resetForm: this.resetForm,
      triggerSave: this.triggerSave$.next.bind(this.triggerSave$),
      triggerLoad: this.triggerLoad$.next.bind(this.triggerLoad$),
      destroy$: this.destroy$,
      loadResource: this.loadResource$.next.bind(this.loadResource$)
    };
  }

}